﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.Windows.Forms;

namespace TestProject
{
    /// <summary>
    /// 그래프 컨트롤
    /// </summary>
    public partial class GraphControl : UserControl
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region Field

        /// <summary>
        /// 텍스트 색상
        /// </summary>
        public Color TextColor = Color.Black;

        /// <summary>
        /// 선 리스트
        /// </summary>
        public List<Line> LineList = new List<Line>();

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region Field

        /// <summary>
        /// 후면 버퍼 비트맵
        /// </summary>
        protected Bitmap backBufferBitmap = null;

        /// <summary>
        /// 전면 버퍼 비트맵
        /// </summary>
        protected Bitmap frontBufferBitmap = null;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 레전드 표시 여부 - ShowLegend

        /// <summary>
        /// 레전드 표시 여부
        /// </summary>
        public bool ShowLegend { get; set; }

        #endregion
        #region 축 표시 여부 - ShowAxis

        /// <summary>
        /// 축 표시 여부
        /// </summary>
        public bool ShowAxis { get; set; }

        #endregion
        #region 그리드 표시 여부 - ShowGrid

        /// <summary>
        /// 그리드 표시 여부
        /// </summary>
        public bool ShowGrid { get; set; }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - GraphControl()

        /// <summary>
        /// 생성자
        /// </summary>
        public GraphControl()
        {
            InitializeComponent();

            ShowLegend = true;
            ShowAxis   = true;
            ShowGrid   = true;

            this.timer.Tick        += timer_Tick;
            this.pictureBox.Resize += pictureBox_Resize;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 타이머 시작하기 - StartTimer()

        /// <summary>
        /// 타이머 시작하기
        /// </summary>
        public void StartTimer()
        {
            this.timer.Enabled = true;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 버퍼 크기 변경하기 - ResizeBuffer()

        /// <summary>
        /// 버퍼 크기 변경하기
        /// </summary>
        protected void ResizeBuffer()
        {
            int width  = this.pictureBox.Width;
            int height = this.pictureBox.Height;

            width  = width  < 10 ? 10 : width;
            height = height < 10 ? 10 : height;

            Bitmap previousBitmap = this.backBufferBitmap;

            if(previousBitmap == null)
            {
                this.backBufferBitmap = new Bitmap(width, height, PixelFormat.Format32bppArgb);
            }
            else
            {
                if((this.backBufferBitmap.Width != width) || (this.backBufferBitmap.Height != height))
                {
                    this.backBufferBitmap = new Bitmap(width, height, PixelFormat.Format32bppArgb);
                }
            }

            if(this.backBufferBitmap != previousBitmap)
            {
                if(previousBitmap != null)
                {
                    previousBitmap.Dispose();
                }
            }
        }

        #endregion
        #region 레전드 그리기 - DrawLegend(graphics, rectangle, legendList, colorList)

        /// <summary>
        /// 레전드 그리기
        /// </summary>
        /// <param name="graphics">그래픽스</param>
        /// <param name="rectangle">사각형</param>
        /// <param name="legendList">레전드 리스트</param>
        /// <param name="colorList">색상 리스트</param>
        protected void DrawLegend(Graphics graphics, Rectangle rectangle, List<string> legendList, List<Color> colorList)
        {
            int legentCount = legendList.Count;

            int rectangleX = rectangle.X;

            for(int i = 0; i < legendList.Count; i++)
            {
                int rectangleWidth = rectangle.X  + (rectangle.Width * (i + 1) / legentCount);

                using(SolidBrush fillBrush = new SolidBrush(colorList[i]))
                {
                    graphics.FillRectangle(fillBrush, rectangleX, rectangle.Y, rectangleWidth, rectangle.Height);
                }

                using(SolidBrush stringBrush = new SolidBrush(colorList[i].GetBrightness() > 0.5 ? Color.Black : Color.White))
                {
                    SizeF legendSize = graphics.MeasureString(legendList[i], Font);

                    PointF legentPoint = new PointF
                    (
                        (float)(0.5 * (rectangleWidth + rectangleX - legendSize.Width)), 
                        (float)(rectangle.Y + 0.5 * (legendSize.Height - legendSize.Height))
                    );

                    graphics.DrawString(legendList[i], Font, stringBrush, legentPoint);
                }

                rectangleX = rectangleWidth;
            }
        }

        #endregion
        #region 그래프 칠하기 - PaintGraph(graphics, width, height)

        /// <summary>
        /// 그래프 칠하기
        /// </summary>
        /// <param name="graphics">그래픽스</param>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        protected void PaintGraph(Graphics graphics, int width, int height)
        {
            double upperX = 0d;
            double lowerX = 0d;
            double upperY = 0d;
            double lowerY = 0d;

            bool foundFirst = false;

            List<string> legendList = new List<string>();
            List<Color>  colorList  = new List<Color>();

            for(int i = 0; i < LineList.Count; i++)
            {
                if(LineList[i] == null)
                {
                    continue;
                }

                if(LineList[i].Visible == false)
                {
                    continue;
                }

                if(!foundFirst)
                {
                    upperX = LineList[i].MaximumX;
                    lowerX = LineList[i].MinimumX;
                    upperY = LineList[i].MaximumY;
                    lowerY = LineList[i].MinimumY;

                    foundFirst = true;
                }

                legendList.Add(LineList[i].Name     );
                colorList.Add (LineList[i].LineColor);

                upperX = upperX > LineList[i].MaximumX ? upperX : LineList[i].MaximumX;
                lowerX = lowerX < LineList[i].MinimumX ? lowerX : LineList[i].MinimumX;
                upperY = upperY > LineList[i].MaximumY ? upperY : LineList[i].MaximumY;
                lowerY = lowerY < LineList[i].MinimumY ? lowerY : LineList[i].MinimumY;
            }

            graphics.Clear(BackColor);

            Rectangle drawRectangle = new Rectangle(0, 0, width, height);

            if(ShowLegend)
            {
                int legendHeight = 5 + (int)graphics.MeasureString("AjtW", Font).Height;

                if(height - legendHeight > 10)
                {
                    drawRectangle = new Rectangle(0, 0, width, height - legendHeight);

                    Rectangle legendRectangle = new Rectangle(0, height - legendHeight + 2, width, legendHeight - 2);

                    DrawLegend(graphics, legendRectangle, legendList, colorList);    
                }
            }

            if(ShowAxis)
            {
                // 추후 추가
            }

            if(!foundFirst)
            {
                return;
            }

            for(int i = 0; i < LineList.Count; i++)
            {
                if(LineList[i] == null)
                {
                    continue;
                }

                if(LineList[i].Visible == false)
                {
                    continue;
                }

                LineList[i].Draw(graphics, drawRectangle, lowerX, upperX, lowerY, upperY);
            }
        }

        #endregion
        #region 버퍼 교체하기 - SwapBuffer()

        /// <summary>
        /// 버퍼 교체하기
        /// </summary>
        protected void SwapBuffer()
        {
            Bitmap temporaryBitmap = this.frontBufferBitmap;

            this.frontBufferBitmap = this.backBufferBitmap;

            this.backBufferBitmap = temporaryBitmap;
        }

        #endregion
        #region 그래프 그리기 - DrawGraph()

        /// <summary>
        /// 그래프 그리기
        /// </summary>
        protected void DrawGraph()
        {
            ResizeBuffer();

            using(Graphics graphics = Graphics.FromImage(this.backBufferBitmap))
            {
                PaintGraph(graphics, this.backBufferBitmap.Width, this.backBufferBitmap.Height);
            }

            SwapBuffer();

            this.pictureBox.Image = this.frontBufferBitmap;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private
        //////////////////////////////////////////////////////////////////////////////// Event

        #region 타이머 틱 처리하기 - timer_Tick(sender, e)

        /// <summary>
        /// 타이머 틱 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void timer_Tick(object sender, EventArgs e)
        {
            DrawGraph();
        }

        #endregion
        #region 픽처 박스 크기 조정시 처리하기 - pictureBox_Resize(sender, e)

        /// <summary>
        /// 픽처 박스 크기 조정시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void pictureBox_Resize(object sender, EventArgs e)
        {
            StartTimer();
        }

        #endregion
    }
}