﻿using Microsoft.DirectX.DirectSound;
using Microsoft.Win32.SafeHandles;
using System;
using System.Collections.Generic;
using System.Threading;

namespace TestProject
{
    /// <summary>
    /// 사운드 소스
    /// </summary>
    public class SoundSource : IDisposable
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Delegate
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 데이터 처리하기 대리자 - ProcessDataDelegate(valueArray)

        /// <summary>
        /// 데이터 처리하기 대리자
        /// </summary>
        /// <param name="valueArray">값 배열</param>
        public delegate void ProcessDataDelegate(float[] valueArray);

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 데이터 처리시 이벤트 - ProcessData

        /// <summary>
        /// 데이터 처리시 이벤트
        /// </summary>
        public event ProcessDataDelegate ProcessData;

        #endregion
        #region FFT 데이터 처리시 이벤트 - ProcessFFTData

        /// <summary>
        /// FFT 데이터 처리시 이벤트
        /// </summary>
        public event ProcessDataDelegate ProcessFFTData;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region Field

        /// <summary>
        /// 리소스 해제 잠금 객체
        /// </summary>
        protected object disposeLock = new object();

        /// <summary>
        /// 리소스 해제 여부
        /// </summary>
        protected bool disposed = false;

        /// <summary>
        /// 캡처
        /// </summary>
        protected Capture capture = null;

        /// <summary>
        /// 캡처 버퍼
        /// </summary>
        protected CaptureBuffer captureBuffer = null;

        /// <summary>
        /// 캡처 스레드
        /// </summary>
        protected Thread captureThread = null;

        /// <summary>
        /// 캡처 버퍼 길이
        /// </summary>
        protected int captureBufferLength = 0;

        /// <summary>
        /// 통지
        /// </summary>
        protected Notify notify;

        /// <summary>
        /// 버퍼 위치 이벤트
        /// </summary>
        protected AutoResetEvent bufferPositionEvent = null;

        /// <summary>
        /// 버퍼 위치 이벤트 핸들
        /// </summary>
        protected SafeWaitHandle bufferPositionEventHandle = null;

        /// <summary>
        /// 스레드 종료 이벤트
        /// </summary>
        protected ManualResetEvent threadTerminateEvent = null;

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 초 단위 버퍼 시간
        /// </summary>
        private const int BUFFER_TIME_IN_SECOND = 5;

        /// <summary>
        /// 초 단위 통지 포인트
        /// </summary>
        private const int NOTIFICATION_POINT_IN_SECOND = 2;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region GUID - GUID

        /// <summary>
        /// GUID
        /// </summary>
        public Guid GUID { get; set; }

        #endregion
        #region 설명 - Description

        /// <summary>
        /// 설명
        /// </summary>
        public string Description { get; set; }

        #endregion
        #region 디폴트 여부 - IsDefault

        /// <summary>
        /// 디폴트 여부
        /// </summary>
        public bool IsDefault
        {
            get
            {
                return GUID == Guid.Empty;
            }
        }

        #endregion
        #region 캡처 여부 - IsCapturing

        /// <summary>
        /// 캡처 여부
        /// </summary>
        public bool IsCapturing { get; protected set; }

        #endregion
        #region 샘플 비율 (단위 : HZ) - SampleRate

        /// <summary>
        /// 샘플 비율 (단위 : HZ)
        /// </summary>
        public int SampleRate { get; set; }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - SoundSource()

        /// <summary>
        /// 생성자
        /// </summary>
        public SoundSource()
        {
            CaptureDevicesCollection collection = new CaptureDevicesCollection();

            for(int i = 0; i < collection.Count; i++)
            {
                if(collection[i].DriverGuid == Guid.Empty)
                {
                    GUID        = collection[i].DriverGuid;
                    Description = collection[i].Description;

                    break;
                }
            }

            this.bufferPositionEvent       = new AutoResetEvent(false);
            this.bufferPositionEventHandle = this.bufferPositionEvent.SafeWaitHandle;
            this.threadTerminateEvent      = new ManualResetEvent(true);
        }

        #endregion
        #region 생성자 - SoundSource(guid, description)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="guid">GUID</param>
        /// <param name="description">설명</param>
        public SoundSource(Guid guid, string description)
        {
            GUID        = guid;
            Description = description;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Destructor

        #region 소멸자 - ~SoundSource()

        /// <summary>
        /// 소멸자
        /// </summary>
        ~SoundSource()
        {
            Dispose(false);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 사운드 소스 리스트 구하기 - GetSoundSourceList()

        /// <summary>
        /// 사운드 소스 리스트 구하기
        /// </summary>
        /// <returns>사운드 소스 리스트</returns>
        public static List<SoundSource> GetSoundSourceList()
        {
            CaptureDevicesCollection collection = new CaptureDevicesCollection();

            List<SoundSource> sourceSourceList = new List<SoundSource>();

            foreach(DeviceInformation deviceInformation in collection)
            {
                sourceSourceList.Add(new SoundSource(deviceInformation.DriverGuid, deviceInformation.Description));
            }

            return sourceSourceList;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 시작하기 - Start()

        /// <summary>
        /// 시작하기
        /// </summary>
        public void Start()
        {
            if(IsCapturing)
            {
                return;
            }

            IsCapturing = true;

            this.capture = new Capture(GUID);

            WaveFormat waveFormat = new WaveFormat();

            waveFormat.Channels              = (short)this.capture.Caps.Channels;
            waveFormat.BitsPerSample         = 16;
            waveFormat.SamplesPerSecond      = 22000;
            waveFormat.FormatTag             = WaveFormatTag.Pcm;
            waveFormat.BlockAlign            = (short)((waveFormat.Channels * waveFormat.BitsPerSample + 7) / 8);
            waveFormat.AverageBytesPerSecond = waveFormat.BlockAlign * waveFormat.SamplesPerSecond;

            this.captureBufferLength = waveFormat.AverageBytesPerSecond * BUFFER_TIME_IN_SECOND;

            CaptureBufferDescription captureBufferDescription = new CaptureBufferDescription();

            captureBufferDescription.Format      = waveFormat;
            captureBufferDescription.BufferBytes = this.captureBufferLength;

            this.captureBuffer = new CaptureBuffer(captureBufferDescription, this.capture);

            int waitHandleCount = BUFFER_TIME_IN_SECOND * NOTIFICATION_POINT_IN_SECOND;

            BufferPositionNotify[] bufferPositionNotifyArray = new BufferPositionNotify[waitHandleCount];

            for(int i = 0; i < waitHandleCount; i++)
            {
                BufferPositionNotify bufferPositionNotify = new BufferPositionNotify();

                bufferPositionNotify.Offset            = (i + 1) * this.captureBufferLength / bufferPositionNotifyArray.Length - 1;
                bufferPositionNotify.EventNotifyHandle = this.bufferPositionEventHandle.DangerousGetHandle();

                bufferPositionNotifyArray[i] = bufferPositionNotify;
            }

            this.notify = new Notify(this.captureBuffer);

            this.notify.SetNotificationPositions(bufferPositionNotifyArray);

            this.threadTerminateEvent.Reset();

            this.captureThread = new Thread(new ThreadStart(ProcessCaptureThread));

            this.captureThread.Name = "Sound capture";

            this.captureThread.Start();
        }

        #endregion
        #region 중단하기 - Stop()

        /// <summary>
        /// 중단하기
        /// </summary>
        public void Stop()
        {
            if(IsCapturing)
            {
                IsCapturing = false;

                this.threadTerminateEvent.Set();

                this.captureThread.Join();

                this.notify.Dispose();

                this.captureBuffer.Dispose();

                this.capture.Dispose();
            }
        }

        #endregion

        #region 리소스 해제하기 - Dispose(disposing)

        /// <summary>
        /// 리소스 해제하기
        /// </summary>
        /// <param name="disposing">리소스 해제 여부</param>
        public void Dispose(bool disposing)
        {
            lock(this.disposeLock)
            {
                if(this.disposed)
                {
                    return;
                }

                this.disposed = true;
            }

            GC.SuppressFinalize(this);

            if(IsCapturing)
            {
                Stop();
            }

            if(this.bufferPositionEventHandle != null)
            {
                this.bufferPositionEventHandle.Dispose();
            }

            if(this.bufferPositionEvent != null)
            {
                this.bufferPositionEvent.Close();
            }

            if(this.threadTerminateEvent != null)
            {
                this.threadTerminateEvent.Close();
            }
        }

        #endregion
        #region 리소스 해제하기 - IDisposable.Dispose()

        /// <summary>
        /// 리소스 해제하기
        /// </summary>
        void IDisposable.Dispose()
        {
            Dispose(true);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Private

        #region 값 처리하기 - ProcessValue(valueArray)

        /// <summary>
        /// 값 처리하기
        /// </summary>
        /// <param name="valueArray">값 배열</param>
        protected void ProcessValue(short[] valueArray)
        {
            ProcessDataDelegate processData    = ProcessData;
            ProcessDataDelegate processFFTData = ProcessFFTData;

            if((processData == null) && (processFFTData == null))
            {
                return;
            }

            int size = 2;

            while(size < valueArray.Length)
            {
                size *= 2;
            }

            float[] xfArray = new float[size];
            float[] xtArray = new float[valueArray.Length];

            for(int i = 0; i < valueArray.Length; i++)
            {
                float value = (float)valueArray[i];

                xfArray[i] = value;
                xtArray[i] = value;
            }

            for(int i = valueArray.Length; i < size; i++)
            {
                xfArray[i] = 0;
            }
       
            if(processData != null)
            {
                processData(xtArray);
            }

            if(processFFTData != null)
            {
                float[] fftArray = new float[xfArray.Length];

                FFT.ComputeForwardFFT(xfArray, fftArray);

                processFFTData(fftArray);
            }
        }

        #endregion
        #region 캡처 스레드 처리하기 - ProcessCaptureThread()

        /// <summary>
        /// 캡처 스레드 처리하기
        /// </summary>
        private void ProcessCaptureThread()
        {
            this.captureBuffer.Start(true);

            try
            {
                int nextCapturePosition = 0;

                WaitHandle[] handleArray = new WaitHandle[] {this.threadTerminateEvent, this.bufferPositionEvent };

                while(WaitHandle.WaitAny(handleArray) > 0)
                {
                    int capturePosition;
                    int readPosition;

                    this.captureBuffer.GetCurrentPosition(out capturePosition, out readPosition);

                    int lockSize = readPosition - nextCapturePosition;

                    if(lockSize < 0)
                    {
                        lockSize += this.captureBufferLength;
                    }

                    if((lockSize & 1) != 0)
                    {
                        lockSize--;
                    }

                    int itemCount = lockSize >> 1;

                    short[] valueArray = (short[])this.captureBuffer.Read
                    (
                        nextCapturePosition,
                        typeof(short),
                        LockFlag.None,
                        itemCount
                    );

                    ProcessValue(valueArray);

                    nextCapturePosition = (nextCapturePosition + lockSize) % this.captureBufferLength;
                }
            }
            catch(Exception exception)
            {
                Console.WriteLine(exception.Message);
            }
            finally
            {
                this.captureBuffer.Stop();
            }
        }

        #endregion
    }
}