﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.Threading;
using System.Windows.Forms;

namespace TestProject
{
    /// <summary>
    /// 사운드 시각화기
    /// </summary>
    public partial class SoundVisualizer : UserControl
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region Field

        /// <summary>
        /// 종료 여부
        /// </summary>
        protected bool closed = false;

        /// <summary>
        /// 후면 버퍼 비트맵
        /// </summary>
        protected Bitmap backBufferBitmap = null;

        /// <summary>
        /// 전면 버퍼 비트맵
        /// </summary>
        protected Bitmap frontBufferBitmap = null;

        /// <summary>
        /// 스택 리스트
        /// </summary>
        protected List<Stack> stackList = new List<Stack>();

        /// <summary>
        /// 스택 리스트 잠금 객체
        /// </summary>
        protected object stackListLockObject = new object();

        /// <summary>
        /// 최소 값
        /// </summary>
        protected float minimumValue = 0;

        /// <summary>
        /// 최대 값
        /// </summary>
        protected float maximumValue = 0;

        /// <summary>
        /// 값 배열 리스트
        /// </summary>
        protected List<float[]> valueArrayList = new List<float[]>();

        /// <summary>
        /// 스레드 잠금 객체
        /// </summary>
        protected object threadLockObject = new object();

        /// <summary>
        /// 작업 스레드
        /// </summary>
        protected Thread workThread = null;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 스펙트럼 너비 - SpectralWidth

        /// <summary>
        /// 스펙트럼 너비
        /// </summary>
        public int SpectralWidth { get; set; }

        #endregion
        #region FFT 데이터 표시 여부 - ShowFFT

        /// <summary>
        /// FFT 데이터 표시 여부
        /// </summary>
        public bool ShowFFT { get; set; }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - SoundVisualizer()
 
        /// <summary>
        /// 생성자
        /// </summary>
        public SoundVisualizer()
        {
            InitializeComponent();

            ShowFFT       = true;
            SpectralWidth = 256;

            Disposed               += UserControl_Disposed;
            this.pictureBox.Resize += pictureBox_Resize;
            this.timer.Tick        += timer_Tick;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 타이머 시작하기 - StartTimer()

        /// <summary>
        /// 타이머 시작하기
        /// </summary>
        public void StartTimer()
        {
            this.timer.Enabled = true;
        }

        #endregion
        #region 데이터 추가하기 - AppendData(valueArray)

        /// <summary>
        /// 데이터 추가하기
        /// </summary>
        /// <param name="valueArray">값</param>
        public void AppendData(float[] valueArray)
        {
            lock(this.threadLockObject)
            {
                this.valueArrayList.Add(valueArray);

                if(this.workThread == null)
                {
                    this.workThread = new Thread(new ThreadStart(ProcessWorkThread));

                    this.workThread.Name = "Sound visualization work thread.";

                    this.workThread.Start();
                }
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Protected
        //////////////////////////////////////////////////////////////////////////////// Event

        #region 사용자 컨트롤 리소스 해제시 처리하기 - UserControl_Disposed(sender, e)

        /// <summary>
        /// 사용자 컨트롤 리소스 해제시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        protected void UserControl_Disposed(object sender, EventArgs e)
        {
            this.closed = true;
        }

        #endregion
        #region 픽처 박스 크기 조정시 처리하기 - pictureBox_Resize(sender, e)

        /// <summary>
        /// 픽처 박스 크기 조정시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        protected void pictureBox_Resize(object sender, EventArgs e)
        {
            this.timer.Enabled = true;
        }

        #endregion
        #region 타이머 틱 처리하기 - timer_Tick(object sender, EventArgs e)

        /// <summary>
        /// 타이머 틱 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void timer_Tick(object sender, EventArgs e)
        {
            Redraw();
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Function

        #region  버퍼 크기 변경하기 - ResizeBuffer()

        /// <summary>
        /// 버퍼 크기 변경하기
        /// </summary>
        protected void ResizeBuffer()
        {
            int width  = this.pictureBox.Width;
            int height = SpectralWidth / 2;

            width  = width  < 10 ? 10 : width;
            height = height < 10 ? 10 : height;
            
            Bitmap previousButmap = this.backBufferBitmap;

            if(previousButmap == null)
            {
                this.backBufferBitmap = new Bitmap(width, height, PixelFormat.Format32bppArgb);
            }
            else
            {
                if((this.backBufferBitmap.Width != width) || (this.backBufferBitmap.Height != height))
                {
                    this.backBufferBitmap = new Bitmap(width, height, PixelFormat.Format32bppArgb);
                }
            }

            if(this.backBufferBitmap != previousButmap)
            {
                if(previousButmap != null)
                {
                    previousButmap.Dispose();
                }
            }
        }

        #endregion
        #region 그래프 칠하기 - PaintGraph(graphics, bitmap, width, height)

        /// <summary>
        /// 그래프 칠하기
        /// </summary>
        /// <param name="graphics"></param>
        /// <param name="bitmap">비트맵</param>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        protected void PaintGraph(Graphics graphics, Bitmap bitmap, int width, int height)
        {
            List<Stack> stackList = new List<Stack>();

            lock(this.stackListLockObject)
            {
                if(this.stackList.Count > 1)
                {
                    stackList.AddRange(this.stackList);

                    this.stackList.Clear();
                }
            }
            
            if(this.frontBufferBitmap == null)
            {
                graphics.Clear(Color.Black);
            }
            else
            {
                graphics.DrawImage(this.frontBufferBitmap, new Point(stackList.Count, 0));

                if(stackList.Count < 1)
                {
                    return;
                }
            }

            for(int i = 0; i < stackList.Count; i++)
            {
                int colorArrayLength = stackList[0].ColorArray.Length; 

                for(int j = 0; j < height; j++)
                {
                    bitmap.SetPixel(i, colorArrayLength - j - 1, stackList[stackList.Count - 1 - i].ColorArray[j]);
                }
            }
        }

        #endregion
        #region 버퍼 교체하기 - SwapBuffer()

        /// <summary>
        /// 버퍼 교체하기
        /// </summary>
        protected void SwapBuffer()
        {
            Bitmap temporaryBitmap = this.frontBufferBitmap;

            this.frontBufferBitmap = this.backBufferBitmap;

            this.backBufferBitmap  = temporaryBitmap;
        }

        #endregion
        #region 다시 그리기 - Redraw()

        /// <summary>
        /// 다시 그리기
        /// </summary>
        protected void Redraw()
        {
            ResizeBuffer();

            using(Graphics graphics = Graphics.FromImage(this.backBufferBitmap))
            {
                PaintGraph(graphics, this.backBufferBitmap, this.backBufferBitmap.Width, this.backBufferBitmap.Height);
            }

            SwapBuffer();

            this.pictureBox.Image = this.frontBufferBitmap;
        }

        #endregion
        #region 스택 리스트 추가하기 - AddStackList(sourceValueArray)

        /// <summary>
        /// 스택 리스트 추가하기
        /// </summary>
        /// <param name="sourceValueArray">소스 값 배열</param>
        protected void AddStackList(float[] sourceValueArray)
        {
            List<Stack> stackList = new List<Stack>();

            int     k                  = 0;
            float[] temporarValueArray = new float[SpectralWidth];
            float[] fftArray           = new float[temporarValueArray.Length];

            int nextK = 0;

            while(k <= sourceValueArray.Length - temporarValueArray.Length)
            {
                nextK = k + (temporarValueArray.Length >> 3);

                for(int i = 0; i < temporarValueArray.Length; i++)
                {
                    temporarValueArray[i] = sourceValueArray[k++];
                }

                FFT.ComputeForwardFFT(temporarValueArray, fftArray);

                float[] targetValueArray = ShowFFT ? fftArray : temporarValueArray;

                float minmiumValue = targetValueArray[0];
                float maximumValue = targetValueArray[0];
              
                for(int i = 0; i < targetValueArray.Length; i++)
                {
                    minmiumValue = minmiumValue < targetValueArray[i] ? minmiumValue : targetValueArray[i];
                    maximumValue = maximumValue > targetValueArray[i] ? maximumValue : targetValueArray[i];
                }

                minmiumValue = float.IsNaN(minmiumValue) ? this.minimumValue : minmiumValue;
                maximumValue = float.IsNaN(maximumValue) ? this.maximumValue : maximumValue;

                minmiumValue = float.IsInfinity(minmiumValue) ? this.minimumValue : minmiumValue;
                maximumValue = float.IsInfinity(maximumValue) ? this.maximumValue : maximumValue;

                this.minimumValue = (255 * this.minimumValue + minmiumValue) / 256;
                this.maximumValue = (255 * this.maximumValue + maximumValue) / 256;

                stackList.Add(new Stack(targetValueArray, this.minimumValue, this.maximumValue));

                k = nextK;
            }

            lock(this.stackListLockObject)
            {
                this.stackList.AddRange(stackList);
            }
        }

        #endregion
        #region 작업 스레드 처리하기 - ProcessWorkThread()

        /// <summary>
        /// 작업 스레드 처리하기
        /// </summary>
        protected void ProcessWorkThread()
        {
            float[] lastValueArray = null;

            while(this.closed == false)
            {
                Thread.Sleep(100);

                float[] valueArray = null;

                lock(this.threadLockObject)
                {
                    if(this.valueArrayList.Count > 0)
                    {
                        valueArray = this.valueArrayList[0];

                        this.valueArrayList.RemoveAt(0);
                    }
                }

                if(valueArray != null)
                {
                    if(lastValueArray != null)
                    {
                        int spectralWidth = SpectralWidth;

                        float[] temporaryValueArray = new float[(int)(spectralWidth * 1.75f)];

                        int k                   = 0; 
                        int swo                 = spectralWidth - (spectralWidth >> 3);
                        int lastValueArrayStart = lastValueArray.Length - swo;

                        lastValueArrayStart = lastValueArrayStart < 0 ? 0 : lastValueArrayStart;

                        for(int i = lastValueArrayStart; i < lastValueArray.Length; i++)
                        {
                            temporaryValueArray[k++] = lastValueArray[i];
                        }
                      
                        for(int i = 0; i < swo && k < temporaryValueArray.Length && i < valueArray.Length; i++)
                        {
                            temporaryValueArray[k++] = valueArray[i];
                        }

                        AddStackList(temporaryValueArray);
                    }

                    AddStackList(valueArray);
                    
                    lastValueArray = valueArray;
                }
            }
        }

        #endregion
    }
}